/* eslint max-depth: ["error", 6] */
import { Uint8ArrayList } from 'uint8arraylist';
import { unsigned } from 'uint8-varint';
import errCode from 'err-code';
// Maximum length of the length section of the message
export const MAX_LENGTH_LENGTH = 8; // Varint.encode(Number.MAX_SAFE_INTEGER).length
// Maximum length of the data section of the message
export const MAX_DATA_LENGTH = 1024 * 1024 * 4;
var ReadMode;
(function (ReadMode) {
    ReadMode[ReadMode["LENGTH"] = 0] = "LENGTH";
    ReadMode[ReadMode["DATA"] = 1] = "DATA";
})(ReadMode || (ReadMode = {}));
const defaultDecoder = (buf) => {
    const length = unsigned.decode(buf);
    defaultDecoder.bytes = unsigned.encodingLength(length);
    return length;
};
defaultDecoder.bytes = 0;
export function decode(options) {
    const decoder = async function* (source) {
        const buffer = new Uint8ArrayList();
        let mode = ReadMode.LENGTH;
        let dataLength = -1;
        const lengthDecoder = options?.lengthDecoder ?? defaultDecoder;
        const maxLengthLength = options?.maxLengthLength ?? MAX_LENGTH_LENGTH;
        const maxDataLength = options?.maxDataLength ?? MAX_DATA_LENGTH;
        for await (const buf of source) {
            buffer.append(buf);
            while (buffer.byteLength > 0) {
                if (mode === ReadMode.LENGTH) {
                    // read length, ignore errors for short reads
                    try {
                        dataLength = lengthDecoder(buffer);
                        if (dataLength < 0) {
                            throw errCode(new Error('invalid message length'), 'ERR_INVALID_MSG_LENGTH');
                        }
                        if (dataLength > maxDataLength) {
                            throw errCode(new Error('message length too long'), 'ERR_MSG_DATA_TOO_LONG');
                        }
                        const dataLengthLength = lengthDecoder.bytes;
                        buffer.consume(dataLengthLength);
                        if (options?.onLength != null) {
                            options.onLength(dataLength);
                        }
                        mode = ReadMode.DATA;
                    }
                    catch (err) {
                        if (err instanceof RangeError) {
                            if (buffer.byteLength > maxLengthLength) {
                                throw errCode(new Error('message length length too long'), 'ERR_MSG_LENGTH_TOO_LONG');
                            }
                            break;
                        }
                        throw err;
                    }
                }
                if (mode === ReadMode.DATA) {
                    if (buffer.byteLength < dataLength) {
                        // not enough data, wait for more
                        break;
                    }
                    const data = buffer.sublist(0, dataLength);
                    if (options?.onData != null) {
                        options.onData(data);
                    }
                    yield data;
                    buffer.consume(dataLength);
                    mode = ReadMode.LENGTH;
                }
            }
        }
        if (buffer.byteLength > 0) {
            throw errCode(new Error('unexpected end of input'), 'ERR_UNEXPECTED_EOF');
        }
    };
    return decoder;
}
/**
 * @param {*} reader
 * @param {import('./types').DecoderOptions} [options]
 * @returns
 */
decode.fromReader = (reader, options) => {
    let byteLength = 1; // Read single byte chunks until the length is known
    const varByteSource = (async function* () {
        while (true) {
            try {
                const { done, value } = await reader.next(byteLength);
                if (done === true) {
                    return;
                }
                if (value != null) {
                    yield value;
                }
            }
            catch (err) {
                if (err.code === 'ERR_UNDER_READ') {
                    return { done: true, value: null };
                }
                throw err;
            }
            finally {
                // Reset the byteLength so we continue to check for varints
                byteLength = 1;
            }
        }
    }());
    /**
     * Once the length has been parsed, read chunk for that length
     */
    const onLength = (l) => { byteLength = l; };
    return decode({
        ...(options ?? {}),
        onLength
    })(varByteSource);
};
//# sourceMappingURL=decode.js.map